/*******************************************************************************
!   Copyright (C) 2009 Intel Corporation. All Rights Reserved.
!
!   The information and  material ("Material") provided below is  owned by Intel
!   Corporation  or its  suppliers  or  licensors, and  title  to such  Material
!   remains with Intel Corporation or  its suppliers or licensors.  The Material
!   contains proprietary  information of Intel  or its suppliers  and licensors.
!   The Material is protected by worldwide copyright laws and treaty provisions.
!   No  part  of  the  Material  may  be  used,  copied,  reproduced,  modified,
!   published, uploaded,  posted, transmitted,  distributed or disclosed  in any
!   way without Intel's prior express  written permission.  No license under any
!   patent, copyright or  other intellectual property rights in  the Material is
!   granted  to  or  conferred  upon  you,  either  expressly,  by  implication,
!   inducement,  estoppel or  otherwise.   Any license  under such  intellectual
!   property rights must be express and approved by Intel in writing.
!******************************************************************************/
using System;
using System.Security;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using mkl;


/**
 * Example showing how to call the Intel MKL BLAS cgemm()
 * function using the CBLAS interface.
 */
public unsafe class test_cgemm
{

    private test_cgemm() {}
    public static void Main(string[] args) {

      /* Data initialization */
      int Order = CBLAS.ORDER.RowMajor;
      int TransA = CBLAS.TRANSPOSE.NoTrans;
      int TransB = CBLAS.TRANSPOSE.NoTrans;

      int M=2, N=4, K=3;
      int lda=K, ldb=N, ldc=N;
      int incx = 1, L = M*K;

      Complex[] A = new Complex[]
      {
        new Complex(1, 2), new Complex( 3,-4), new Complex( 3,3),
        new Complex(1,-7), new Complex(11,1),  new Complex(-1,2)
      };

      Complex[] B = new Complex[]
      {
        new Complex(-4, 3), new Complex( 3,9),  new Complex(7,3), new Complex(-3, 6),
        new Complex( 4,-9), new Complex(-7,10), new Complex(4,1), new Complex( 1,-8),
        new Complex( 1,-8), new Complex(-8,1),  new Complex(6,4), new Complex(-3,-2)
      };

      Complex[] C = new Complex[]
      {
        new Complex(-7,-69), new Complex(-23,52), new Complex(23, 34), new Complex(-47,-43),
        new Complex(85,-54), new Complex(-15,74), new Complex(57,-23), new Complex( 65,-64)
      };

      Complex alpha = new Complex(1,1);
      Complex beta  = new Complex(-1,-1);

    	Console.WriteLine("MKL cblas_cgemm example");
  		Console.WriteLine();
      /* Print initial data */
      Console.Write(String.Format("alpha  = ({0,3}, {1,3})\n", alpha.re,  alpha.im));
      Console.Write(String.Format("beta   = ({0,3}, {1,3})\n",  beta.re,   beta.im));
      Console.WriteLine();
      printComplexMatrix("Matrix A",A,M,K);
      printComplexMatrix("Matrix B",B,K,N);
      printComplexMatrix("Initial C",C,M,N);

      /* Computation */
      CBLAS.cgemm(Order, TransA, TransB, M, N, K,
          alpha, A, lda, B, ldb, beta, C, ldc);

      /* Print the result */
      printComplexMatrix("Resulting C after CGEMM",C,M,N);
      Console.WriteLine("TEST PASSED");
  		Console.WriteLine();

	}

    /** Print the matrix X assuming row-major order of elements. */
    private static void printComplexMatrix(String prompt, Complex[] A, int I, int J) {
        Console.WriteLine(prompt);
        for (int i=0; i<I; i++) {
            for (int j=0; j<J; j++)
                Console.Write(String.Format("\t({0,3}, {1,3})",A[i*J+j].re, A[i*J+j].im));
            Console.WriteLine();
        }
    }
}

namespace mkl
{
  [StructLayout(LayoutKind.Sequential,CharSet=CharSet.Ansi)]
  public unsafe struct Complex
  {
    public float re, im;

    public Complex(float re, float im)
    {
      this.re = re;
      this.im = im;
    }
  }

	/** CBLAS wrappers */
	public unsafe sealed class CBLAS
	{
		private CBLAS() {}
		/** Constants for CBLAS_ORDER enum, file "mkl_cblas.h" */
		public sealed class ORDER
		{
			private ORDER() {}
			public static int RowMajor=101;  /* row-major arrays */
			public static int ColMajor=102;  /* column-major arrays */
		}

		/** Constants for CBLAS_TRANSPOSE enum, file "mkl_cblas.h" */
		public sealed class TRANSPOSE
		{
			private TRANSPOSE() {}
			public static int NoTrans  =111; /* trans='N' */
			public static int Trans    =112; /* trans='T' */
			public static int ConjTrans=113; /* trans='C' */
		}

		/** CBLAS cblas_cgemm wrapper */
		public static void cgemm(int Order, int TransA, int TransB,
			int M, int N, int K, Complex alpha, Complex[] A, int lda,
			Complex[] B, int ldb, Complex beta, Complex[] C, int ldc)
		{
      fixed (Complex* pA = &A[0])
      fixed (Complex* pB = &B[0])
      fixed (Complex* pC = &C[0])
			CBLASNative.cblas_cgemm(Order, TransA, TransB, M, N, K,
				&alpha, pA, lda, pB, ldb, &beta, pC, ldc);
		}

	}

	/** CBLAS native declarations */
	[SuppressUnmanagedCodeSecurity]
	internal sealed unsafe class CBLASNative
	{
		private CBLASNative() {}

		/** CBLAS native cblas_cgemm declaration */
		[DllImport("mkl_rt.dll", CallingConvention=CallingConvention.Cdecl,
			 ExactSpelling=true, SetLastError=false)]
		internal static extern void cblas_cgemm(
			int Order, int TransA, int TransB, int M, int N, int K,
			Complex *alpha, Complex* A, int lda, Complex* B, int ldb,
			Complex* beta, Complex* C, int ldc);

  }

}
